"use strict";
/// <reference path="../node_modules/types.d.ts/modules/typed-controls/index.d.ts" />
//no/ <reference path="../node_modules/types.d.ts/modules/dialog-promise/index.d.ts" />
/// <reference path="../node_modules/types.d.ts/modules/codenautas-umd/index.d.ts" />
/* No sabemos cómo referenciar a los tipos escritos en mwOwn/index.d.ts */ 
//NO/ <reference path="../node_modules/types.d.ts/modules/myOwn/index.d.ts" />
/// <reference path="../node_modules/types.d.ts/modules/myOwn/in-myOwn.d.ts" />
/// <reference path="./in-backend-plus.d.ts" />
/// <amd-dependency path="../node_modules/types.d.ts/modules/typed-controls/index.d.ts" name="TypedControls" />
/// <amd-dependency path="../node_modules/types.d.ts/modules/dialog-promise/index.d.ts" name="DialogPromise" />
/// <amd-dependency path="../node_modules/types.d.ts/modules/type-store/index.d.ts" name="TypeStore" />

/*jshint eqnull:true */
/*jshint node:true */

(function codenautasModuleDefinition(root, name, factory) {
    /* global define */
    /* istanbul ignore next */
    if(typeof root.globalModuleName !== 'string'){
        root.globalModuleName = name;
    }
    /* istanbul ignore next */
    if(typeof exports === 'object' && typeof module === 'object'){
        module.exports = factory();
    }else if(typeof define === 'function' && define.amd){
        define(factory);
    }else if(typeof exports === 'object'){
        // @ts-ignore
        exports[root.globalModuleName] = factory();
    }else{
        root[root.globalModuleName] = factory();
    }
    root.globalModuleName = null;
})(/*jshint -W040 */this, 'myOwn', function() {
/*jshint +W040 */

/*jshint -W004 */
var myOwn = {};
/*jshint +W004 */

// var DialogPromise = require('dialog-promise');
// var TypeStore = require('type-store');
// var TypedControls = require('typed-controls');

myOwn.messages={};
myOwn.i18n={messages:{}};

myOwn.i18n.messages.en = {
    autoCloseWhenEnds: "auto close when ends",
    Cancel   : "Cancel",
    notLogged: "Not logged in",
    noServer : "The server is inaccessible",
    noNetwork: "Not connected to the network",
    reLogin  : "Sign in",
    stopCountDown: "stop count down"
};

myOwn.i18n.messages.es = {
    autoCloseWhenEnds: "cerrar la ventana al terminar",
    Cancel   : "Cancelar",
    notLogged: "Sin sesión activa",
    noServer : "No se puede acceder al servidor",
    noNetwork: "No se detecta conexión de red",
    reLogin  : "Iniciar sesión",
    stopCountDown: "detener la cuenta regresiva"
};

var bestGlobals = require('best-globals');

var coalesce = bestGlobals.coalesce;
var changing = bestGlobals.changing;

var jsYaml = require('js-yaml');
var TypeStore=require('type-store');

var JSON4all = require('json4all');
var LocalDbTransaction = require('./my-localdb').LocalDbTransaction;
var LocalDb = require('./my-localdb').LocalDb;
var WebSqlDb = require('./my-websqldb').WebsqlDb;

/** @param {T} x
 *  @returns {T}
 *  @template {T}
 */
function id(x){return x;}

/** @param {string} name 
  * @param {number|undefined} version
  * @returns {Promise<IDBDatabase>}
  */
function idbOpen(name, version){
    /** @type {IDBRequest} */
    return IDBX(indexedDB.open(name, version));
}

/** @param {IDBRequest} request */
function IDBX(request){
    return new Promise(function(resolve, reject){
        if('onsuccess' in request){
            request.onsuccess=function(event){
                resolve(request.result);
            }
        }else{
            request.oncomplete=function(event){
                resolve(request.result);
            }
        }
        request.onerror=function(reject){
            alertPromise(request.error.message)
            reject(request.error);
        }
    })
}

if(typeof window != "undefined"){
    window.IDBX = IDBX;
}

myOwn.statusDivName = 'reconnection_div';

myOwn.autoSetupFunctions = [
    function autoSetupMyThings(){
        var my = this;
        if(!('server' in my)){
            my.server = {connected: null, broadcaster:html.div({id:'server-status-broadcaster'}).create()}
        }
        var readProcedureDefinitions = function readProcedureDefinitions(){
            var promise;
            if(!my.offline.mode){
                promise = my.ajaxPromise({
                    action:'client-setup',
                    method:'get',
                    encoding:'JSON',
                    parameters:[],
                    progress:false
                })
            }else{
                promise = Promise.resolve(
                    JSON.parse(localStorage.getItem('setup'))
                );
            }
            return promise.then(function(setup){
                localStorage.setItem('setup', JSON.stringify(setup));
                my.config = setup;
            }).then(function(){
                my.config.procedure=my.config.procedure||{};
                my.config.procedures.forEach(/** @param {bp.ProcedureDef} procedureDef */function(procedureDef){
                    my.config.procedure[procedureDef.action]=procedureDef;
                    var target;
                    /** @type {string} */
                    var lastName;
                    procedureDef.parameter=procedureDef.parameter||{};
                    procedureDef.parameters.forEach(function(parameterDef){
                        procedureDef.parameter[parameterDef.name]=parameterDef;
                    });
                    var partsNames=procedureDef.action.split('/').filter(id).forEach(function(name){
                        if(lastName){
                            if(!target[lastName]){
                                target[lastName]={"dont-use":lastName};
                            }else if(target[lastName]["dont-use"]!==lastName){
                                throw new Error("Bad nesting of procedures");
                            }
                            target=target[lastName];
                        }else{
                            target=my.ajax;
                        }
                        lastName=name;
                    });
                    target[lastName]=my.ajaxPromise.bind(my,procedureDef);
                    var activeUserElement = document.getElementById('active-user');
                    if(activeUserElement){
                        activeUserElement.textContent=my.config.username||'-';
                    }
                });
                DialogPromise.path.img=my.path.img;
                TypedControls.path.img=my.path.img;
                var gridBuffer = my.config.config['grid-buffer'];
                var offlineMode = gridBuffer && (gridBuffer === 'idbx' || gridBuffer === 'wsql');
                if(offlineMode){
                    try{
                        switch(gridBuffer) {
                            case 'idbx':
                              my.ldb = new LocalDb(my.appName+my.clientVersion);
                              ///** @type {(callback:(ldb:any)=>Promise<T>)=>Promise<T>} */
                              //my.inLdb = new LocalDbTransaction(my.appName+my.clientVersion).getBindedInTransaction()
                              break;
                            case 'wsql':
                              my.ldb = new WebSqlDb(my.appName+my.clientVersion);
                              break;
                            default:
                              throw new Error('grid buffer name is bad defined')
                          }
                    }catch(err){
                        my.log(err);
                    }
                }
            });
        };
        my.captureKeys();
        // document.addEventListener('click', function(event){
        //     if(document.buttonLupa && !document.buttonLupa.forElement.contains(event.target)){
        //         my.quitarLupa();
        //     }
        // });
        setInterval(function(){
            my.testKeepAlive();
        },2000);
        window.addEventListener("error", function myErrorHandler(error, url, lineNumber) {
            my.log('error', error.message);
        });
        window.addEventListener("unhandledrejection", function(event) {
            if(!event.reason.DialogPromise){
                my.log('error', event.reason.message);
            }
        });
        TypeStore.options.doNotCopyNonCopyables=true;
        TypeStore.options.doNotOutputNonCopyables=true;
        if(!('offline' in my)){
            my.offline={};
            Object.defineProperty(my.offline, 'mode', {
                get:function(){
                    return localStorage.getItem(my.appName+my.clientVersion+'_offline_mode')=='offline';
                },
                set:function(offline){
                    return localStorage.setItem(my.appName+my.clientVersion+'_offline_mode', offline?'offline':'online');
                }
            })
        }
        return readProcedureDefinitions();
    }
];
myOwn.deleteLocalDb = function deleteLocalDb(){
    var gridBuffer = my.config.config['grid-buffer'];
    var offlineMode = gridBuffer && (gridBuffer === 'idbx' || gridBuffer === 'wsql');
    if(offlineMode){
        try{
            switch(gridBuffer) {
                case 'idbx':
                  return LocalDb.deleteDatabase(my.appName+my.clientVersion);              
                case 'wsql':
                  return WebSqlDb.deleteDatabase(my.appName+my.clientVersion);              
                default:
                  throw new Error('grid buffer name is bad defined')
              }
        }catch(err){
            my.log(err);
        }
    }
    
}
myOwn.autoSetup = function autoSetup(){
    var my=this;
    if(my.promiseChainAutoSetup){
        console.log('Multiple calls to autoSetup()');
        return my.promiseChainAutoSetup;
    }
    var promiseChain=Promise.resolve();
    my.autoSetupFunctions.forEach(function(autoSetupFunction){
        promiseChain=promiseChain.then(autoSetupFunction.bind(my));
    });
    my.promiseChainAutoSetup=promiseChain;
    return promiseChain;
}
    
myOwn["log-severities"] = {
    error:{permanent:true },
    log  :{permanent:false},
};
    
myOwn.log = function log(severity, message){
    var my=this;
    var consoleMessage;
    var clientMessage;
    if(!message){
        if(severity instanceof Error){
            clientMessage=severity.message;
            consoleMessage=severity;
            severity='error';
        }else{
            clientMessage=severity;
            severity='log';
        }
    }else{
        clientMessage = message;
    }
    consoleMessage = consoleMessage || clientMessage;
    var status = document.getElementById('mini-console');
    console.log(severity, consoleMessage);
    status = status||document.body;
    var divMessage = html.div({"class": ["status-"+severity]}, clientMessage).create();
    if(status.firstChild){
        status.insertBefore(divMessage, status.firstChild);
    }else{
        status.appendChild(divMessage);
    }
    if(!this["log-severities"][severity].permanent){
        setTimeout(function(){
            my.fade(divMessage);
        },3000);
    }else{
        setTimeout(function(){
            my.fade(divMessage);
        },30000);
    }
};

myOwn.fade = function fade(element, options){
    if(element.tagName.toUpperCase()==='TR' && element.parentNode.replaceChild){
        var parent=element.parentNode;
        var dummyTr=document.createElement(element.tagName);
        dummyTr.className=element.className;
        options=options||{};
        options.smooth=options.smooth||{};
        var spans=options.smooth.spans;
        var sourceRow;
        if(spans){
            sourceRow=spans.map(function(span){
                return {tagName:'td', colSpan:span};
            })
        }else{
            sourceRow=Array.prototype.slice.call(element.cells);
        }
        sourceRow.forEach(function(cell){
            var dummyCell=document.createElement(cell.tagName);
            dummyTr.appendChild(dummyCell);
            dummyCell.className=cell.className;
            dummyCell.colSpan=cell.colSpan;
        });
        var div=document.createElement('div');
        dummyTr.cells[1].appendChild(div);
        div.style.height=Math.min(window.innerHeight*0.5, element.cells[0].offsetHeight-2)+'px';
        div.style.transition='height 0.6s ease';
        div.style.overflowY='hidden';
        parent.replaceChild(dummyTr, element);
        if(options.smooth.content){
            div.appendChild(options.smooth.content);
        }
        setTimeout(function(){
            div.style.height='1px';
        },10);
        setTimeout(function(){
            parent.removeChild(dummyTr);
        },500);
    }else{
        element.parentNode.removeChild(element);
    }
};


myOwn.focusFirstColumnOf = function focusFirstColumnOf(row){
    var my = this;
    var element = my.nextElementThatIs(row.cells[0],my.beInteractive);
    element.focus();
}


myOwn.insertRow = function insertRow(where){
    var my = this;
    var section, iRow, newTr;
    if(where.under){
        section = where.under.parentNode;
        iRow = where.under.sectionRowIndex+1;
    }else if('iRow' in where && (where.table||where.section)){
        section = where.table||where.section;
        iRow = where.iRow;
    }else{
        throw new Error('my-tables insert without good where');
    }
    var tr = section.insertRow(iRow);
    if(where.smooth){
        if(where.smooth===true){ 
            where.smooth={};
        }
        var trDummy = section.insertRow(iRow);
        tr.style.display='none';
        for(var i=0; i<(where.smooth.colCount||1); i++){
            var cell=trDummy.insertCell(-1);
            cell.colSpan=(where.smooth.spans||{})[i]||1;
        }
        var divDummy=html.div({class:'grid-dummy-cell-ini'}).create();
        trDummy.appendChild(divDummy);
        setTimeout(function(){
            divDummy.style.height=(where.smooth.height||20)+'px';
        },10);
        setTimeout(function(){
            trDummy.style.display='none';
            tr.style.display='';
            section.removeChild(trDummy);
            if(where.autoFocus!==false){
                setTimeout(function(){
                    my.focusFirstColumnOf(tr);
                    window.keyStarForceFocusNow=true;
                },100);
            }
        },500);
    }
    return tr;
};

myOwn.adaptData = function adaptData(tableDef, rows){
    rows.forEach(function(row){
        tableDef.fields.forEach(function(fieldDef){
            if(fieldDef.typeName=='number' && fieldDef.exact && fieldDef.decimals){
                if(row[fieldDef.name]){
                    row[fieldDef.name] = row[fieldDef.name]-0;
                }
            }
            /*
            if(fieldDef.typeName=='date'){
                if(row[fieldDef.name] && !(row[fieldDef.name] instanceof Date)){
                    row[fieldDef.name] = bestGlobals.date.iso(row[fieldDef.name]);
                }
            }
            */
        });
    });
};

Object.defineProperty(myOwn, 'ajax', {
    get: function ajax(){
        if(!this.config){
            throw new Error("before use myOwn.ajax, myOwn.autoSetup() must be called");
        }else{
            return this.ajaxPromise;
        }
    }
});

myOwn.encoders = {
    JSON: { parse: JSON.parse, stringify: JSON.stringify },
    plain: { 
        parse: function(x){return x;}, 
        stringify: function(x){
            if(typeof x === "object" /* && !(x instanceof FileList)*/){
                throw new Error("Invalid plain type "+(x==null?value:typeof x)+" detected");
            }
            return x;
        }
    },
    yaml: {
        parse: jsYaml.load.bind(jsYaml),
        stringify: jsYaml.dump.bind(jsYaml),
    },
    JSON4all: {
        parse: JSON4all.parse,
        stringify: JSON4all.stringify,
    },
};

myOwn.launcherOk=function(launcher){
    return function(result){
        launcher.title=launcher.originalTitle;
        launcher.removeAttribute('my-working');
        return result;
    }
}

myOwn.launcherCatch=function(launcher){
    return function(err){
        launcher.title=err.message;
        launcher.setAttribute('my-working','error');
        throw err;
    }
}

myOwn.propertiesWhiteList=['message','detail','code','table','constraint','name','severity'];

myOwn.alertError = function(err){
    return dialogPromise(function(dialogWindow, closeWindow){
        var button=html.button(DialogPromise.messages.Ok).create();
        button.addEventListener('click',function(){
            closeWindow();
        });
        dialogWindow.appendChild(html.table([html.tr([
            html.td({class:'dialog-error-img'}, [html.img({src:my.path.img+'warning128.png'})]),
            html.td([
                html.pre(myOwn.propertiesWhiteList.map(function(name){ 
                    return err[name]?err[name]+'\n':'';
                }).join('')),
                html.div([button])
            ])
        ])]).create());
    });
}

myOwn.ajaxPromise = function ajaxPromise(procedureDef,data,opts){
    opts = opts || {};
    if(!('visiblyLogErrors' in opts)){
        opts.visiblyLogErrors=true;
    }
    var my = this;
    if(opts.launcher){
        if(!opts.launcher.originalTitle){
            opts.launcher.originalTitle=opts.launcher.title;
        }else{
            opts.launcher.title=opts.launcher.originalTitle;
        }
        opts.launcher.setAttribute('my-working','working');
    }
    return Promise.resolve().then(function(){
        var params={};
        procedureDef.parameters.forEach(function(paramDef){
            var value=coalesce(data[paramDef.name],'defaultValue' in paramDef?paramDef.defaultValue:coalesce.throwErrorIfUndefined("lack of parameter "+paramDef.name));
            value = my.encoders[paramDef.encoding].stringify(value);
            params[paramDef.name]=value;
        });
        if(data && data.files){
            params.files=data.files;
        }
        var result=[];
        var progress=procedureDef.progress!==false;
        var divProgress;
        var onClose=function(){}
        var defaultInformProgress = function defaultInformProgress(progressInfo){
            if(progressInfo.message){
                if(!divProgress){
                    var idAutoClose='id-auto-close-'+Math.random();
                    var checkAutoClose=html.input({type:'checkbox', id:idAutoClose, checked:myOwn.ajaxPromise.autoClose}).create();
                    var divButton=html.div({style:'height:40px'},[
                        my.messages.viewProgress,
                        html.div([
                            checkAutoClose,
                            html.label({for:idAutoClose}, my.messages.autoCloseWhenEnds),
                        ])
                    ]).create();
                    checkAutoClose.addEventListener('change',function(){
                        myOwn.ajaxPromise.autoClose=checkAutoClose.checked;
                    });
                    divProgress=html.div({class:'result-progress'}).create();
                    simpleFormPromise({elementsList:[divButton,divProgress]});
                    onClose=function(err){
                        var closeButton=html.button(DialogPromise.messages.Ok).create();
                        var stopCountDown=html.button(my.messages.stopCountDown).create();
                        divButton.innerHTML="";
                        divButton.appendChild(closeButton);
                        closeButton.onclick=function(){
                            divButton.dialogPromiseDone();
                        }
                        if(err){
                            divProgress.appendChild(html.div({class:'error-message'},err.message).create())
                        }
                        if(myOwn.ajaxPromise.autoClose && !err){
                            setTimeout(function(){
                                divButton.dialogPromiseDone();
                            },myOwn.ajaxPromise.autoClose);
                        }
                    }
                }
                divProgress.insertBefore(
                    html.div({class:'my-progress'},progressInfo.message).create(),
                    divProgress.childNodes[0]
                );
            }
        }
        var informProgress;
        if(opts.divProgress){
            divProgress=opts.divProgress;
            if(opts.informProgress){
                informProgress = function(progress){
                    if(progress.message){
                        defaultInformProgress(progress)
                    }else{
                        opts.informProgress(progress)
                    }
                }
            }else{
                informProgress = defaultInformProgress;
            }
        }else{
            informProgress = opts.informProgress || defaultInformProgress;
        }
        var controlLoggedIn = function controlLoggedIn(result){
            if(result && result[0]=="<" && result.match(/login/m)){
                my.informDetectedStatus('notLogged');
                throw changing(new Error(my.messages.notLogged),{displayed:true, isNotLoggedError:true});
            }
            my.informDetectedStatus('logged', true);
        }
        return AjaxBestPromise[procedureDef.method]({
            multipart:procedureDef.files,
            url:procedureDef.action,
            data:params,
            uploading:opts.uploading
        }).onLine(function(line,ender){
            controlLoggedIn(line);
            if(progress){
                if(line.substr(0,2)=='--'){
                    progress=false;
                }else{
                    var info=JSON.parse(line);
                    if(info.error){
                        var err = new Error(info.error.message);
                        likeAr(info.error).forEach(function(value, name){
                            if(name!='message'){
                                err[name]=value;
                            }
                        })
                        throw err;
                    }
                    informProgress(info.progress);
                }
            }else{
                result.push(line||ender);
            }
        }).then(function(){
            onClose();
            result=result.join('');
            controlLoggedIn(result);
            return my.encoders[procedureDef.encoding].parse(result);
        }).catch(function(err){
            onClose(err);
            if(opts.launcher){
                opts.launcher.setAttribute('my-working','error');
                opts.launcher.title='err';
            }
            if(!err.isNotLoggedError) {
                if(!window.navigator.onLine) {
                    my.informDetectedStatus('noNetwork');
                }else if(!!err.originalError) {
                    my.informDetectedStatus('noServer');
                    if(my.server.connected){
                        my.server.connected = false
                        my.server.broadcaster.dispatchEvent(new Event('noServer'));
                    }
                } else {
                    my.informDetectedStatus('logged', true);
                }
            }
            if(!err.displayed && opts.visiblyLogErrors || err.status==403){
                my.log(err);
            }
            throw err;
        });
    });
};

myOwn.ajaxPromise.autoClose=6000;

myOwn.testKeepAlive = function testKeepAlive(){
    var my = this;
    var element = document.getElementById('keep-alive-signal') || my.debugging && document.body.appendChild(html.div({id:'keep-alive-signal'}).create());
    if(element){
        element.textContent='t';
        element.style.backgroundColor='#4DD';
        element.style.display='';
    }
    var startTime=new Date().getTime();
    return my.ajaxPromise({
        parameters:[],
        method:'post',
        action:'keep-alive.json',
        encoding:'plain',
        progress:false
    },{},{visiblyLogErrors:false}).then(function(){
        if(window.updateOnlineStatus){
            updateOnlineStatus();
        }
        var lightServer = document.getElementById('light-server');
        if(lightServer){
            lightServer.src=my.path.img+'server-ok.png';
            if(!my.server.connected){
                my.server.connected = true;
                my.server.broadcaster.dispatchEvent(new Event('serverConnected'));
            }
            var speed=1000/(1+new Date().getTime()-startTime);
            if(isNaN(lightServer.result.speed)){
                lightServer.result.speed = speed; 
            }
            lightServer.result.speed = Math.ceil((lightServer.result.speed*7 + speed)/8); 
            lightServer.title = lightServer.result.speed;
        }
    }).then(function(){
        if(element){
            element.textContent='ok';
            element.style.backgroundColor='lightgreen';
        }
    },function(){
        if(element){
            element.textContent='E!';
            element.style.backgroundColor='#F88';
        }
    }).then(function(){
        if(element){
            setTimeout(function(){
                element.style.display='none';
            }, 2000);
        }
    })
};

myOwn.showQuestion = function showQuestion(message, opts){
    return confirmPromise(message, opts);
};

myOwn.easeInOut = function easeInOut(currentTime, start, change, duration) {
    currentTime /= duration / 2;
    if (currentTime < 1) {
        return change / 2 * currentTime * currentTime + start;
    }
    currentTime -= 1;
    return -change / 2 * (currentTime * (currentTime - 2) - 1) + start;
};

myOwn.scrollToTop = function(element, to, duration) {
    var start = element.scrollTop,
        change = to - start,
        increment = 20;
    var me = this;
    var animateScroll = function(elapsedTime) {        
        elapsedTime += increment;
        var position = me.easeInOut(elapsedTime, start, change, duration);                        
        element.scrollTop = position; 
        if (elapsedTime < duration) {
            setTimeout(function() {
                animateScroll(elapsedTime);
            }, increment);
        } else {
            element.scrollTop = document.documentElement.scrollTop = 0; // cross browser scrolling to top
        }
    };
    animateScroll(0);
};

myOwn["connection-status"]={
   logged   : { show: false, mustAsk: false },
   notLogged: { show: true , mustAsk: {idMessage:'reLogin', url:'login'}},
   noServer : { show: true , mustAsk: false, id:'light-server'        , img:'server-error'     },
   noNetwork: { show: true , mustAsk: false, id:'light-network-signal', img:'network-no-signal'},
};

myOwn.debuggingStatus=false;  // /* 
if(new Date()<bestGlobals.datetime.ymdHms(2016,10,25,14,50,0)){
    myOwn.debuggingStatus=function(statusCode){
        myOwn.debuggingStatus.count=(myOwn.debuggingStatus.count||0)+1
        if(!window.debuggingStatusDiv){
            var box=document.createElement('div');
            box.style.position='fixed';
            box.style.left='400px';
            box.style.top='30px';
            box.style.border='1px dashed green';
            box.style.backgroundColor='rgba(230,230,180,0.5)';
            box.textContent='actual status:';
            window.debuggingStatusDiv=document.createElement('div');
            document.body.appendChild(box);
            box.appendChild(window.debuggingStatusDiv);
            window.debuggingStatusDiv.id='debuggingStatusDiv';
        }
        window.debuggingStatusDiv.textContent=statusCode+' '+myOwn.debuggingStatus.count;
    };
}

myOwn.informDetectedStatus = function informDetectedStatus(statusCode, logged) {
    if(!logged){
        window.location.href='login';
    }
};

function isInteracive(element){
"use strict";
    return (
        (element instanceof HTMLInputElement || element instanceof HTMLButtonElement || element instanceof HTMLTextAreaElement)
        && !element.disabled
        || element instanceof HTMLElement && element.contentEditable=="true"
    )
    && element.style.display!='none'
    && element.style.visibility!='hidden'
    && !element.getAttribute('skip-enter')
    && (element.tabIndex==null || element.tabIndex>=0);
}

function nextElement(elemento, noGoDownWhen){
"use strict";
    var proximo=elemento;
    var no_me_voy_a_colgar=200;
    if(elemento.children.length && !noGoDownWhen(proximo)){
        while(proximo.children.length>0 && no_me_voy_a_colgar-- && !noGoDownWhen(proximo)){
            proximo=proximo.children[0];
        }
        return proximo;
    }else{
        while(proximo && proximo.nextElementSibling===null && no_me_voy_a_colgar--){
            proximo=proximo.parentNode;
        }
        if(proximo){
            return proximo.nextElementSibling;
        }
    }
    return null;
}

function proximo_elemento_que_sea(elemento, controlador){
"use strict";
    var proximo=nextElement(elemento,controlador);
    var no_me_voy_a_colgar=2000;
    while(proximo && !controlador(proximo) && no_me_voy_a_colgar--){
        proximo=nextElement(proximo,controlador);
    }
    return proximo;
}

myOwn.beInteractive = function beInteractive(element, controller){
    return isInteracive(element)
}

myOwn.nextElementThatIs = function nextElementThatIs(element, controller){
    return proximo_elemento_que_sea(element, controller);
}

function enter_hace_tab_en_este_elemento(elemento){
    return elemento.tagName!="TEXTAREA";
}

function tableInfo(element){
    var info={};
    while(element && element.tagName!='TABLE'){
        if(element.tagName=='TR'){
            info.tr=info.tr||element;
        }
        if(element.tagName=='TD'){
            info.td=info.td||element;
        }
        element=element.parentNode;
    }
    if(element){
        info.table=element;
    }
    return info;
}

function getCaretPosition(element) {
    // https://stackoverflow.com/questions/4811822/get-a-ranges-start-and-end-offsets-relative-to-its-parent-container/4812022#4812022
    var caretOffset = 0;
    var doc = element.ownerDocument || element.document;
    var win = doc.defaultView || doc.parentWindow;
    var sel;
    if (typeof win.getSelection != "undefined") {
        sel = win.getSelection();
        if (sel.rangeCount > 0) {
            var range = win.getSelection().getRangeAt(0);
            var preCaretRange = range.cloneRange();
            preCaretRange.selectNodeContents(element);
            preCaretRange.setEnd(range.endContainer, range.endOffset);
            caretOffset = preCaretRange.toString().length;
        }
    } else if ( (sel = doc.selection) && sel.type != "Control") {
        var textRange = sel.createRange();
        var preCaretTextRange = doc.body.createTextRange();
        preCaretTextRange.moveToElementText(element);
        preCaretTextRange.setEndPoint("EndToEnd", textRange);
        caretOffset = preCaretTextRange.text.length;
    }
    return caretOffset;
}

myOwn.captureKeys = function captureKeys() {
    var previousKey;
    var previousPosition;
    var my=this;
    if(this.captureKeysInstaled){
        throw new Error('captureKeysInstaled');
    }
    this.captureKeysInstaled=true;
    document.addEventListener('keydown', function(evento){
        // document.title=evento.which+','+evento.shiftKey+','+evento.ctrlKey+','+evento.altKey+','+evento.metaKey;
        if(evento.which==37 && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // KeyLeft
            var info=tableInfo(this.activeElement);
            if(info.table){
                if(evento.ctrlKey || info.td.textContent=='' || !getCaretPosition(info.td)){
                    var newPos=info.td.cellIndex-1;
                    while(newPos && !my.beInteractive(info.tr.cells[newPos])){
                        newPos--;
                    }
                    if(my.beInteractive(info.tr.cells[newPos])){
                        info.tr.cells[newPos].focus();
                        if(document.activeElement){
                            var sel = window.getSelection()
                            sel.collapse(document.activeElement,document.activeElement.childNodes.length);
                        }
                        evento.preventDefault();
                    }
                }
            }
        }
        var goRight=function(){
            var newPos=info.td.cellIndex+1;
            while(newPos<=info.tr.cells.length && !my.beInteractive(info.tr.cells[newPos])){
                newPos++;
            }
            if(my.beInteractive(info.tr.cells[newPos])){
                info.tr.cells[newPos].focus();
                evento.preventDefault();
            }
        };
        if(evento.which==39 && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // KeyRight
            var info=tableInfo(this.activeElement);
            if(info.table){
                if(evento.ctrlKey || info.td.textContent=='' || (previousKey == 39 && getCaretPosition(info.td)==info.td.textContent.length)){
                    goRight();
                    previousPosition=false;
                }else{
                    previousPosition=getCaretPosition(info.td)
                }
            }
        }
        if((evento.which==40 || evento.which==38)  && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // KeyDown, KeyUp
            var info=tableInfo(this.activeElement);
        }
        if((evento.which==40 || evento.which==38)  && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // KeyDown, KeyUp
            var info=tableInfo(this.activeElement);
            if(info.table){
                var newPos=info.tr.rowIndex+evento.which-39;
                var newRow=info.table.rows[newPos];
                if(newRow){
                    var newCell = newRow.cells[info.td.cellIndex];
                    if(my.beInteractive(newCell)){
                        newCell.focus();
                        evento.preventDefault();
                    }
                }
            }
        }
        if(evento.which==115 && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // F4
            var info=tableInfo(this.activeElement);
            if(info.table){
                var abovePos=info.tr.rowIndex-1;
                var aboveRow=info.table.rows[abovePos];
                if(aboveRow){
                    var aboveCell = aboveRow.cells[info.td.cellIndex];
                    if(aboveCell.getTypedValue){
                        var value=aboveCell.getTypedValue();
                        if(info.td.setTypedValue){
                            info.td.setTypedValue(value, true);
                            var belowPos=info.tr.rowIndex+1;
                            var belowRow=info.table.rows[belowPos];
                            if(belowRow){
                                var belowCell = belowRow.cells[info.td.cellIndex];
                                if(my.beInteractive(belowCell)){
                                    belowCell.focus();
                                    evento.preventDefault();
                                }
                            }
                        }
                    }
                }
            }
        }
        if(evento.which==115 && evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){ // F4
            var info=tableInfo(this.activeElement);
            if(info.table){
                var abovePos=info.tr.rowIndex-1;
                var aboveRow=info.table.rows[abovePos];
                if(aboveRow){
                    var aboveCell = aboveRow.cells[info.td.cellIndex];
                    if(aboveCell.getTypedValue){
                        var value=aboveCell.getTypedValue();
                        if(info.td.setTypedValue){
                            info.td.setTypedValue(value, true);
                            goRight();
                        }
                    }
                }
            }
        }
        previousKey=evento.which;
    });
    document.addEventListener('keypress', function(evento){
        if(my.config.useragent.isiPad && evento.which==231){ // F4
            var info=tableInfo(this.activeElement);
            if(info.table){
                var abovePos=info.tr.rowIndex-1;
                var aboveRow=info.table.rows[abovePos];
                if(aboveRow){
                    var aboveCell = aboveRow.cells[info.td.cellIndex];
                    if(aboveCell.getTypedValue){
                        var value=aboveCell.getTypedValue();
                        if(info.td.setTypedValue){
                            info.td.setTypedValue(value, true);
                            var belowPos=info.tr.rowIndex+1;
                            var belowRow=info.table.rows[belowPos];
                            if(belowRow){
                                var belowCell = belowRow.cells[info.td.cellIndex];
                                if(my.beInteractive(belowCell)){
                                    belowCell.focus();
                                    evento.preventDefault();
                                }
                            }
                        }
                    }
                }
            }
        }
        if(window.keyStarForceFocusNow){
            window.keyStarForceFocusNow=false;
            if(evento.which==106 && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey){
                this.activeElement.focus();
            }
        }
        if(evento.which==13 && !evento.shiftKey  && !evento.ctrlKey  && !evento.altKey  && !evento.metaKey && !this.activeElement.getAttribute("enter-clicks")){ // Enter
            var enfoco=this.activeElement;
            var este=this.activeElement;
            if(enter_hace_tab_en_este_elemento(este)){
                var no_me_voy_a_colgar=2000;
                while(este && this.activeElement===enfoco && no_me_voy_a_colgar--){
                    este=proximo_elemento_que_sea(este,isInteracive);
                    if(este){
                        este.focus();
                    }
                }
                if(este){
                    evento.preventDefault();
                }
            }
        }
    });
};

myOwn.getRect = function getRect(element){
    var rect = {top:0, left:0, width:element.offsetWidth, height:element.offsetHeight};
    while( element != null ) {
        rect.top += element.offsetTop;
        rect.left += element.offsetLeft;
        element = element.offsetParent;
    }
    return rect;
}

// myOwn.quitarLupa = function quitarLupa(){
//     if(document.buttonLupa){
//         document.body.removeChild(document.buttonLupa);
//         clearTimeout(document.buttonLupaTimmer);
//         document.buttonLupaTimmer=null;
//         document.buttonLupa=null;
//     }
// }

myOwn.path={
    img:'img/'
}

myOwn.getUniqueDomId = function getUniqueDomId(){
    myOwn.getUniqueDomId.serial=(myOwn.getUniqueDomId.serial||0)+1;
    return 'uniQ$___$id'+myOwn.getUniqueDomId.serial;
}

myOwn.prepareFloating3dots = function prepareFloating3dots(){
    if(!my.imgFloating3dots && false){
        var img = html.img({src:my.path.img+'floating-3dots.png', class:'floating-img'}).create();
        my.imgFloating3dots=img;
        var rePosition = function(){
            img.style.visibility=window.scrollY || window.scrollX?'visible':'hidden';
            img.style.top = window.innerHeight - 40 + 'px';
            img.style.left = window.innerWidth - 40 + 'px';
        }
        img.style.position='fixed';
        document.body.appendChild(img);
        setTimeout(rePosition,100);
        window.addEventListener('scroll', rePosition);
        window.addEventListener('resize', rePosition);
    }
}

myOwn.inlineCss = function inlineCss(id){
    var autoStyle = document.getElementById(id);
    if(!autoStyle){
        var autoStyle=document.createElement('style');
        autoStyle.type = 'text/css';
        autoStyle.id=id;
        document.getElementsByTagName('head')[0].appendChild(autoStyle);
    }
    return autoStyle;
}

myOwn.prepareRulerToggle = function prepareRulerToggle(){
    if(!my.imgRulerToggle){
        var img = html.img({src:my.path.img+'floating-ruler-toggle.png', class:'floating-img'}).create();
        my.imgRulerToggle=img;
        var rePosition = function(){
            img.style.visibility=window.scrollY || window.scrollX?'visible':'hidden';
            img.style.top = '-10px';
            img.style.left = '-10px';
        }
        img.style.position='fixed';
        document.body.appendChild(img);
        setTimeout(rePosition,100);
        window.addEventListener('scroll', rePosition);
        window.addEventListener('resize', rePosition);
        var floatingStyle="position:relative; background-color:rgba(155,155,255,0.8) !important; background-clip: padding-box; outline:2px solid black;";
        var autoPosition=function(){
            var tables = document.querySelectorAll(".my-grid");
            if(tables){
                Array.prototype.forEach.call(tables,function(table){
                    var myTableName = table.getAttribute("my-table");
                    var autoStyleTop = my.inlineCss("css-my-table-"+myTableName+"-top");
                    var rect = my.getRect(table.tHead?table.tHead.rows[0]:table);
                    var cssClausules=[];
                    autoStyleTop.actualDif=my.autoRuler && window.scrollY-rect.top;
                    // var cssSelectorPrefix="[my-table=\""+myTableName.replace(/"/g,'\\"')+"\"] > ";
                    var cssSelectorPrefix="[my-table="+JSON.stringify(myTableName)+"] > ";
                    if(!('previousDif' in autoStyleTop) || autoStyleTop.actualDif!=autoStyleTop.previousDif){
                        if(autoStyleTop.actualDif>0){
                            cssClausules.push(
                                cssSelectorPrefix + "thead > tr > th { "+floatingStyle+
                                " top:"+(autoStyleTop.actualDif)+'px; '+
                                "}"
                            );
                        }
                        autoStyleTop.previousDif = autoStyleTop.actualDif;
                        autoStyleTop.innerHTML=cssClausules.join('\n');
                    }
                    var autoStyleLeft = my.inlineCss("css-my-table-"+myTableName+"-left");
                    var cssSelectorH = cssSelectorPrefix + "thead > tr > [my-fixed2left-column]";
                    var cssSelectorB = cssSelectorPrefix + "tbody > tr > [my-fixed2left-column]";
                    var firstCell=Array.prototype.slice.call(table.querySelectorAll(cssSelectorH),0,100).find(function(cell){
                        return cell.scrollWidth>4;
                    });
                    if(firstCell){
                        var rect = my.getRect(firstCell);
                        var cssClausules=[];
                        autoStyleLeft.actualDif=my.autoRuler && window.scrollX-rect.left;
                        if(!('previousDif' in autoStyleLeft) || autoStyleLeft.actualDif!=autoStyleLeft.previousDif){
                            if(autoStyleLeft.actualDif>0){
                                cssClausules.push(
                                    cssSelectorB+
                                    " { "+floatingStyle+
                                    " left:"+(autoStyleLeft.actualDif)+'px; '+
                                    "}"
                                );
                            }
                            autoStyleLeft.previousDif = autoStyleLeft.actualDif;
                            autoStyleLeft.innerHTML=cssClausules.join('\n');
                        }
                    }
                });
            }
        }
        img.onclick=function(){
            my.autoRuler=!my.autoRuler;
            my.imgRulerToggle.src=my.path.img+'floating-ruler-toggle-'+(my.autoRuler?'on':'off')+'.png'
            setTimeout(function(){
                autoPosition();
            },100);
        }
        window.addEventListener('scroll', autoPosition);
        window.addEventListener('resize', autoPosition);
    }
}

myOwn.createSmartButton = function createSmartButton(opts){
    var button=opts.button || html.button(opts.buttonProperties||{}, opts.buttonContent||opts.buttonLabel).create();
    Object.defineProperty(button, 'smartState', {
        get:function(){ return this.getAttribute('smart-button'); },
        set:function(value){ 
            this.disabled=value!='active';
            this.setAttribute('smart-button',value); 
        }
    })
    button.smartState=opts.initialState||'unkown';
    button.onclick=function(){
        var button=this;
        button.smartState='working';
        Promise.resolve().then(function(){
            if(opts.confirmMessage){
                return confirmPromise(opts.confirmMessage);
            }
        }).then(function(){
            return opts.mainFun({ajaxOpts:{launcher:button}});
        }).then(function(result){
            button.smartState='unknown';
            (opts.okFun||function(result){
                button.smartState='active';
                return alertPromise(result,{reject:false});
            })(result);
        },function(err){
            button.smartState='error';
            (opts.errorFun||function(err){
                button.smartState='active';
                return alertPromise(err.message,{reject:false});
            })(err);
        })
    };
    if(opts.insideElement){
        opts.insideElement.appendChild(button);
    }
    return button;
};


return myOwn;

});
